//-------------------------------------------------------------------
// File    : ub_xnet_l1.h
// Author  : Uwe Becker
// Date    : 2018-01-14
//
// Project : XNET
// Modul   : Protocol layer L1 [communication layer]
// Node-To-Node connection
//-------------------------------------------------------------------

#ifndef UB_XNET_L1_H_
#define UB_XNET_L1_H_

//-------------------------------------------------------------------
// don't make changes below this line
// to port this project to another MCU or hardware:
// change "xnet_l0.c", "xnet_l0.h", "xnet_global.h"
//-------------------------------------------------------------------

// include xnet layer-0 and global header
#include "ub_xnet_global.h"
#include "ub_xnet_l0.h"

// version number for XNET layer-1 (8bit)
#define XNET_VERSION_L1_8BIT			0x03	// 0x03 = version 0.3


#define L2_FLAG_NONE					0x0000	// no flag raised
#define L2_FLAG_OWN_LINK_STATE_CHANGED	0x0001	// own LinkState changed
#define L2_FLAG_FLOOD_OWN_LINK_STATE	0x0002	// flood own LinkState
#define L2_FLAG_CALC_NEW_PATH_TABLE		0x0004	// calculate new PathTable
#define L2_FLAG_FLOOD_COMPLETE_LS_TABLE	0x0008	// flood complete LinkStateTable
#define L2_FLAG_REQ_COMPLETE_LS_TABLE	0x0010	// request the complete LinkStateTable


// debug ----------------------------------------------
#define DBG_LIST_SIZE		20

#define DBG_LIST_RX_FRAME		0x01 // L1
#define DBG_LIST_TX_FRAME		0x02
#define DBG_LIST_UPD_LS			0x03
#define DBG_LIST_UPD_LST		0x11 // L2
#define DBG_LIST_CALC_PT		0x12
//-----------------------------------------------------


// enumeration for TX-State
typedef enum {
	TX_OK = 0,
	TX_DATA_ERROR,
	TX_PORT_DISABLED,
	TX_PACKAGE_SKIPPED,
	TX_NO_RESPONSE,
	TX_COLLISION,
	TX_ERROR
}TX_State_t;

// enumeration for RX-State
typedef enum {
	RX_OK = 0,
	RX_NO_RESPONSE,
	RX_ERROR
}RX_State_t;

// enumeration for Port-Mask
typedef enum {
	PORTMASK_SINGLE_PORT = 0,
	PORTMASK_SINGLE_PORT_HELLO,
	PORTMASK_ANY_PORT,
	PORTMASK_ANY_OTHER_PORT
}Portmask_Mode_t;

// enumeration for NodeState
typedef enum {
	NODE_STATE_INIT = 0,
	NODE_STATE_HELLO,
	NODE_STATE_CONNECTION,
	NODE_STATE_IDLE
}NodeState_t;

// payload strukture
typedef struct {
	U8_CHAR port_nr;
	U8_CHAR frame_id;
	U8_CHAR host_id;
	U8_CHAR data_length;
	U8_CHAR data[XNET_MAX_FRAME_LENGTH];
}Payload_t;

// TX-Frame structure
typedef struct {
	U8_CHAR port_mask;	 					// port_mask
	U8_CHAR frame[XNET_MAX_FRAME_LENGTH];	// frame data
}TX_Frame_t;

// RX-Frame structure
typedef struct {
	U8_CHAR port_nr;	 					// port_nr
	U8_CHAR frame[XNET_MAX_FRAME_LENGTH];	// frame data
}RX_Frame_t;

// TX-Framebuffer
typedef struct {
	U8_CHAR tx_rd_ptr;
	U8_CHAR tx_wr_ptr;
	TX_Frame_t tx_frame[NODE_TX_BUFFER_SIZE];
}TX_Framebuf_t;

// RX-Framebuffer
typedef struct {
	U8_CHAR rx_rd_ptr;
	U8_CHAR rx_wr_ptr;
	RX_Frame_t rx_frame[NODE_RX_BUFFER_SIZE];
}RX_Framebuf_t;

// DEBUG!!
typedef struct {
	U8_CHAR wr_ptr;
	U32_DWORD time[DBG_LIST_SIZE];
	U8_CHAR list[DBG_LIST_SIZE][10];
}DbgCmdList_t;

// DEBUG!!
typedef struct {
	U32_DWORD tx_err_cnt;		// error counter of TX frames
	U32_DWORD rx_err_cnt;		// error counter of RX frames
	U32_DWORD tx_ovf_cnt;		// error counter of TX buffer overflow
	U32_DWORD rx_ovf_cnt;		// error counter of RX buffer overflow
	U32_DWORD buf_ovf_cnt;		// error counter of L2 RX buffer overflow
	U32_DWORD update_ls;
	U32_DWORD calc_PT;
	U32_DWORD rec_ls;
	U32_DWORD collision_cnt;	// collision counter
}DbgCnt_t;

// structure of a single port
typedef struct {
	U8_BOOL enable;				// port enable/disable
	U8_BOOL tx_ok;				// package send to neighbor
	U8_BOOL rx_ok;				// package received from neighbor
	U8_BOOL connected;			// connected or not connected
	U8_CHAR neighbor_id;		// neighbor ID, when port connected
	U32_DWORD hello_cnt;		// hello counter
	U32_DWORD hello_cnt_delay;	// delay value for hello counter
	U8_BOOL new_neighbor;
}Port_t;

// L2-Flags
typedef struct {
	U32_DWORD flags;			// flags for layer L2
	U8_CHAR OWN_LINK_STATE_CHANGED_cnt;
	U8_CHAR CALC_NEW_PATH_TABLE_cnt;
}L2Flags_t;

// counter
typedef struct {
	U32_DWORD loop_cnt;			// counter for loop function
	U32_DWORD delay_cnt;		// counter for delay
	U32_DWORD state_cnt;
	U32_DWORD time_cnt;
}Counter_t;

// structure of the node
typedef struct {
	NodeState_t state;
	U8_CHAR my_id;				// unique ID in the network for this node
	Counter_t cnt;
	DbgCnt_t dbg_cnt;
	L2Flags_t l2_flags;
	TX_Framebuf_t tx_framebuf;	// tx frame buffer
	RX_Framebuf_t rx_framebuf;	// rx frame buffer
	Payload_t tx_payload;		// tx payload
	Payload_t rx_payload;		// rx payload
	Port_t port[NODE_PORT_CNT];	// struct of all ports on this node
	DbgCmdList_t dbg_list;
}Node_t;
Node_t Node;

// function calls for layer-2
void UB_Xnet_L1_Init(U8_CHAR own_id);
void UB_Xnet_L1_1ms_Tick(void);
void UB_Xnet_L1_Loop(void);
U8_CHAR UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(U8_CHAR port_nr, Portmask_Mode_t mode);
U8_BOOL UB_Xnet_L1_Add_Tx_Package(U8_CHAR port_mask, U8_CHAR frame_id);
U8_BOOL UB_Xnet_L1_Check_Rx_Buffer(void);
U8_CHAR UB_Xnet_L1_Rx_Buf_Cnt(void);
U8_BOOL UB_Xnet_L1_Tx_Buffer_Full(void);
U8_CHAR UB_Xnet_L1_Tx_Buf_Cnt(void);

// debug
void UB_Xnet_L1_AddDbgLst(U8_CHAR d0, U8_CHAR d1, U8_CHAR d2, U8_CHAR d3, U8_CHAR d4, U8_CHAR d5, U8_CHAR d6, U8_CHAR d7, U8_CHAR d8, U8_CHAR d9);
void UB_Xnet_L1_ClrDbgLst(void);


#endif /* UB_XNET_L1_H_ */
