//-------------------------------------------------------------------
// File    : ub_xnet_l1.c
// Author  : Uwe Becker
// Date    : 2018-01-14
// Version : 0.01
//
// Project : XNET
// Modul   : Protocol layer L1 [communication layer]
// Node-To-Node connection
//-------------------------------------------------------------------



//-------------------------------------------------------------------
// don't make changes below this line
// to port this project to another MCU or hardware:
// change "xnet_l0.c", "xnet_l0.h", "xnet_global.h"
//-------------------------------------------------------------------


// include own header file
#include "ub_xnet_l1.h"


// internal TX functions
TX_State_t send_package(U8_CHAR port_nr);
TX_State_t tx_package(U8_CHAR port_nr);
TX_State_t tx_open(U8_CHAR port_nr);
TX_State_t tx_frame(U8_CHAR port_nr);
void tx_close(U8_CHAR port_nr);
TX_State_t tx_check_neighbor_response(U8_CHAR port_nr);


// internal RX functions
RX_State_t rx_package(U8_CHAR port_nr);
RX_State_t rx_open(U8_CHAR port_nr);
RX_State_t rx_frame(U8_CHAR port_nr);
void rx_close(U8_CHAR port_nr);
U8_BOOL rx_check_frame(U8_CHAR idx);
RX_State_t rx_process_frame(U8_CHAR port_nr);


// other internal functions
U8_BOOL calc_checksum(U8_CHAR *data_ptr, U8_CHAR *crc_hi, U8_CHAR *crc_lo);
void check_all_ports_for_rx_requests(void);
void check_tx_buffer_and_send(void);
void change_tx_connection_state(U8_CHAR port_nr, U8_BOOL new_state);
void change_rx_connection_state(U8_CHAR port_nr, U8_BOOL new_state);
U8_CHAR get_port_nr_from_tx_port_mask(void);
U8_CHAR get_tx_port_mask(void);
void inc_tx_rd_ptr(void);
void inc_tx_wr_ptr(void);
void inc_rx_rd_ptr(void);
void inc_rx_wr_ptr(void);
U8_CHAR get_rx_port_nr(void);
U8_CHAR get_tx_frame_id(void);




// init xnet layer-1
void UB_Xnet_L1_Init(U8_CHAR own_id)
{
	U8_CHAR n, m;

	// init structure
	Node.state = NODE_STATE_INIT;
	Node.my_id = own_id;

	Node.cnt.delay_cnt = 0;
	Node.cnt.loop_cnt = 0;
	Node.cnt.state_cnt = 0;
	Node.cnt.time_cnt = 0;

	Node.dbg_cnt.tx_err_cnt = 0;
	Node.dbg_cnt.rx_err_cnt = 0;
	Node.dbg_cnt.tx_ovf_cnt = 0;
	Node.dbg_cnt.rx_ovf_cnt = 0;
	Node.dbg_cnt.buf_ovf_cnt = 0;
	Node.dbg_cnt.update_ls = 0;
	Node.dbg_cnt.calc_PT = 0;
	Node.dbg_cnt.rec_ls = 0;
	Node.dbg_cnt.collision_cnt = 0;

	Node.l2_flags.flags = L2_FLAG_NONE;
	Node.l2_flags.OWN_LINK_STATE_CHANGED_cnt = 0;
	Node.l2_flags.CALC_NEW_PATH_TABLE_cnt = 0;

	// TX framebuf
	Node.tx_framebuf.tx_wr_ptr = 0;
	Node.tx_framebuf.tx_rd_ptr = 0;
	for(n = 0; n < NODE_TX_BUFFER_SIZE; n++) {
		Node.tx_framebuf.tx_frame[n].port_mask = 0x00;
		for(m = 0; m < XNET_MAX_FRAME_LENGTH; m++) {
			Node.tx_framebuf.tx_frame[n].frame[m] = 0x00;
		}
	}

	// RX framebuf
	Node.rx_framebuf.rx_wr_ptr = 0;
	Node.rx_framebuf.rx_rd_ptr = 0;
	for(n = 0; n < NODE_RX_BUFFER_SIZE; n++) {
		Node.rx_framebuf.rx_frame[n].port_nr = 0;
		for(m = 0; m < XNET_MAX_FRAME_LENGTH; m++) {
			Node.rx_framebuf.rx_frame[n].frame[m] = 0x00;
		}
	}

	// TX payload
	Node.tx_payload.frame_id = XNET_FRAME_UNDEFINED;
	Node.tx_payload.port_nr = XNET_PORT_NR_UNDEFINED;
	Node.tx_payload.host_id = XNET_NODE_ID_UNDEFINED;
	Node.tx_payload.data_length = 0x00;
	for(m = 0; m < XNET_MAX_FRAME_LENGTH; m++) {
		Node.tx_payload.data[m] = 0x00;
	}

	// RX payload
	Node.rx_payload.frame_id = XNET_FRAME_UNDEFINED;
	Node.rx_payload.port_nr = XNET_PORT_NR_UNDEFINED;
	Node.rx_payload.host_id = XNET_NODE_ID_UNDEFINED;
	Node.rx_payload.data_length = 0x00;
	for(m = 0; m < XNET_MAX_FRAME_LENGTH; m++) {
		Node.rx_payload.data[m] = 0x00;
	}

	for(n = 0; n < NODE_PORT_CNT; n++) {
		Node.port[n].enable = XTRUE;
		Node.port[n].tx_ok = XFALSE;
		Node.port[n].rx_ok = XFALSE;
		Node.port[n].connected = XFALSE;
		Node.port[n].neighbor_id = XNET_NODE_ID_UNDEFINED;
		Node.port[n].hello_cnt = (n * NODE_HELLO_OFFSET);
		Node.port[n].hello_cnt_delay = XNET_HELLO_DELAY_INTERVAL_ms;
		Node.port[n].new_neighbor = XFALSE;
	}

	UB_Xnet_L1_ClrDbgLst();

	// init xnet layer-0
	UB_Xnet_L0_Init();
}

// 1ms Tick
void UB_Xnet_L1_1ms_Tick(void)
{
	U8_CHAR n;

	Node.cnt.loop_cnt++;
	Node.cnt.delay_cnt++;
	Node.cnt.state_cnt++;
	Node.cnt.time_cnt++;

	for(n = 0; n < NODE_PORT_CNT; n++) {
		Node.port[n].hello_cnt++;
	}
}

// udate the node-state
void update_state(void)
{
	static U8_CHAR status_led_mode = 0;

	switch(Node.state) {
		case NODE_STATE_INIT:
			if(Node.cnt.state_cnt > Node.my_id) {
				Node.cnt.state_cnt = 0;
				Node.state = NODE_STATE_HELLO;
			}
		break;
		case NODE_STATE_HELLO:
			if(Node.cnt.state_cnt > XNET_HELLO_PHASE_DURATION) {
				Node.cnt.state_cnt = 0;
				Node.state = NODE_STATE_CONNECTION;
			}
		break;
		case NODE_STATE_CONNECTION:
			if(Node.cnt.state_cnt > XNET_CONNECTION_PHASE_DURATION) {
				Node.cnt.state_cnt = 0;
				Node.state = NODE_STATE_IDLE;
			}
		break;
		case NODE_STATE_IDLE:
			if(status_led_mode < 200) {
				if(Node.cnt.state_cnt > XNET_STATUS_LED_ON_TIME_ms) {
					Node.cnt.state_cnt = 0;
					// status LED
					UB_Xnet_L0_Status_Led_Off();
					status_led_mode++;
				}
			}
			else {
				// status LED
				UB_Xnet_L0_Status_Led_On();
				status_led_mode = 0;
				Node.cnt.state_cnt = 0;
			}
		break;
	}
}

// main level-1 loop
// (must be call periodically from level-2)
void UB_Xnet_L1_Loop(void)
{
	U8_CHAR port_mask;
	static U8_CHAR hello_port = 0;

	update_state();

	if(Node.state == NODE_STATE_INIT) return;

	// check all ports to receive packages
	check_all_ports_for_rx_requests();

	// check if time for hello package on this port
	if(Node.port[hello_port].hello_cnt >= Node.port[hello_port].hello_cnt_delay) {
		// wait until tx buffer is not full
		if(UB_Xnet_L1_Tx_Buffer_Full() == XFALSE) {
			// add hello package for this ports to tx framebuf
			Node.tx_payload.host_id = Node.my_id; // source
			Node.tx_payload.data_length = 0;
			port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(hello_port, PORTMASK_SINGLE_PORT_HELLO);
			UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_HELLO);
			Node.port[hello_port].hello_cnt = 0;
			Node.port[hello_port].hello_cnt_delay = XNET_HELLO_DELAY_INTERVAL_ms;
		}
	}

	hello_port++;
	if(hello_port >= NODE_PORT_CNT) {
		hello_port = 0;
	}

	// check if anything to send and send a single package if so
	check_tx_buffer_and_send();
}

// converts a port_nr into a port_mask
U8_CHAR UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(U8_CHAR port_nr, Portmask_Mode_t mode)
{
	U8_CHAR port_mask = 0, n;

	// toDo: don't flood back to the same neighbor

	if(port_nr >= NODE_PORT_CNT) return 0;

	if(mode == PORTMASK_SINGLE_PORT) {
		// set single port in port_mask
		if((Node.port[port_nr].enable == XTRUE) && (Node.port[port_nr].connected == XTRUE)) {
			port_mask = (1 << port_nr);
		}
		return port_mask;
	}
	else if(mode == PORTMASK_SINGLE_PORT_HELLO) {
		// set single port in port_mask (also if port is not connected yet)
		if(Node.port[port_nr].enable == XTRUE) {
			port_mask = (1 << port_nr);
		}
		return port_mask;
	}
	else if(mode == PORTMASK_ANY_PORT) {
		// set all possible ports on this node
		for(n = 0; n < NODE_PORT_CNT; n++) {
			if((Node.port[n].enable == XTRUE) && (Node.port[n].connected == XTRUE)) {
				port_mask |= (1 << n);
			}
		}
		return port_mask;
	}
	else if(mode == PORTMASK_ANY_OTHER_PORT) {
		// set all possible ports on this node accept the port_nr
		for(n = 0; n < NODE_PORT_CNT; n++) {
			if((Node.port[n].enable == XTRUE) && (Node.port[n].connected == XTRUE) && (n != port_nr)) {
				port_mask |= (1 << n);
			}
		}
		return port_mask;
	}

	return 0;
}

// add a package to the tx framebuf
// return:
// XTRUE = package copied in tx frambuf
U8_BOOL UB_Xnet_L1_Add_Tx_Package(U8_CHAR port_mask, U8_CHAR frame_id)
{
	U8_CHAR idx, n;
	U8_CHAR crc_hi,crc_lo;

	if(port_mask == 0x00) return XFALSE; // empty port mask
	if(Node.tx_payload.data_length > (XNET_MAX_FRAME_LENGTH-6)) return XFALSE;

	idx = Node.tx_framebuf.tx_wr_ptr;
	Node.tx_framebuf.tx_frame[idx].port_mask = port_mask;

	// create header of the package
	Node.tx_framebuf.tx_frame[idx].frame[0] = 6+Node.tx_payload.data_length; // length of frame
	Node.tx_framebuf.tx_frame[idx].frame[1] = frame_id; // frame id
	Node.tx_framebuf.tx_frame[idx].frame[2] = Node.my_id; // own id
	Node.tx_framebuf.tx_frame[idx].frame[3] = Node.tx_payload.host_id; // host id

	// create payload of the package
	if(Node.tx_payload.data_length > 0) {
		for(n = 0; n < Node.tx_payload.data_length; n++) {
			Node.tx_framebuf.tx_frame[idx].frame[4+n] = Node.tx_payload.data[n];
		}
	}
	if(calc_checksum(Node.tx_framebuf.tx_frame[idx].frame, &crc_hi, &crc_lo) == XFALSE) return XFALSE;
	Node.tx_framebuf.tx_frame[idx].frame[4+Node.tx_payload.data_length] = crc_hi;
	Node.tx_framebuf.tx_frame[idx].frame[5+Node.tx_payload.data_length] = crc_lo;

	// fill debug list
	if(frame_id != XNET_FRAME_HELLO) {
		UB_Xnet_L1_AddDbgLst(DBG_LIST_TX_FRAME, frame_id, Node.my_id, port_mask, Node.tx_payload.host_id, Node.tx_payload.data[0], Node.tx_payload.data[1], Node.tx_payload.data[2], Node.tx_payload.data[3], Node.tx_payload.data[4]);
	}

	// new frame in tx buffer, increment pointer
	inc_tx_wr_ptr();

	return XTRUE;
}

// check if a package is in rx buffer
// process the package and fill rx_payload
// return:
// XTRUE = a package from rx buffer processed
U8_BOOL UB_Xnet_L1_Check_Rx_Buffer(void)
{
	U8_BOOL package_received = XFALSE;
	U8_CHAR port_nr;

	// check if something in rx framebuffer
	if(UB_Xnet_L1_Rx_Buf_Cnt() > 0) {
		// get receiving port
		port_nr = get_rx_port_nr();
		if(port_nr < NODE_PORT_CNT) {
			// process received frame
			if(rx_process_frame(port_nr) == RX_OK) {
				package_received = XTRUE;
				if(Node.rx_payload.frame_id != XNET_FRAME_HELLO) {
					// status LED
					if(Node.state == NODE_STATE_IDLE) Node.cnt.state_cnt = 0;
					UB_Xnet_L0_Status_Led_On();
				}
			}
		}
		// switch to next frame in rx framebuffer
		inc_rx_rd_ptr();
	}

	return package_received;
}

// check if TX buffer is full
// return:
// XTRUE = tx buffer is full
U8_BOOL UB_Xnet_L1_Tx_Buffer_Full(void)
{
	if(UB_Xnet_L1_Tx_Buf_Cnt() >= (NODE_TX_BUFFER_SIZE-1)) {
		return XTRUE;
	}

	return XFALSE;
}

// returns number of packages in tx framebuf (0 = empty)
U8_CHAR UB_Xnet_L1_Tx_Buf_Cnt(void)
{
	U8_CHAR cnt = 0;

	if(Node.tx_framebuf.tx_wr_ptr >= Node.tx_framebuf.tx_rd_ptr) {
		cnt = Node.tx_framebuf.tx_wr_ptr - Node.tx_framebuf.tx_rd_ptr;
	}
	else {
		cnt = (Node.tx_framebuf.tx_wr_ptr + NODE_TX_BUFFER_SIZE) - Node.tx_framebuf.tx_rd_ptr;
	}

	return cnt;
}

// add entry in debug list
void UB_Xnet_L1_AddDbgLst(U8_CHAR d0, U8_CHAR d1, U8_CHAR d2, U8_CHAR d3, U8_CHAR d4, U8_CHAR d5, U8_CHAR d6, U8_CHAR d7, U8_CHAR d8, U8_CHAR d9)
{
	Node.dbg_list.time[Node.dbg_list.wr_ptr] = Node.cnt.time_cnt;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][0] = d0;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][1] = d1;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][2] = d2;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][3] = d3;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][4] = d4;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][5] = d5;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][6] = d6;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][7] = d7;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][8] = d8;
	Node.dbg_list.list[Node.dbg_list.wr_ptr][9] = d9;
	Node.dbg_list.wr_ptr++;
	if(Node.dbg_list.wr_ptr >= DBG_LIST_SIZE) Node.dbg_list.wr_ptr = 0;
}

// clear debug list
void UB_Xnet_L1_ClrDbgLst(void)
{
	U8_CHAR n;

	for(n = 0; n < DBG_LIST_SIZE; n++) {
		Node.dbg_list.time[n] = 0;
		Node.dbg_list.list[n][0] = 0;
		Node.dbg_list.list[n][1] = 0;
		Node.dbg_list.list[n][2] = 0;
		Node.dbg_list.list[n][3] = 0;
		Node.dbg_list.list[n][4] = 0;
		Node.dbg_list.list[n][5] = 0;
		Node.dbg_list.list[n][6] = 0;
		Node.dbg_list.list[n][7] = 0;
		Node.dbg_list.list[n][8] = 0;
		Node.dbg_list.list[n][9] = 0;
	}
	Node.dbg_list.wr_ptr = 0;
}

// returns port_mask of current package in tx buffer
U8_CHAR get_tx_port_mask(void)
{
	U8_CHAR idx, port_mask;

	idx = Node.tx_framebuf.tx_rd_ptr;
	port_mask = Node.tx_framebuf.tx_frame[idx].port_mask;

	return port_mask;
}

// returns Frame-ID of current package in tx buffer
U8_CHAR get_tx_frame_id(void)
{
	U8_CHAR idx, frame_id;

	idx = Node.tx_framebuf.tx_rd_ptr;
	frame_id = Node.tx_framebuf.tx_frame[idx].frame[1];

	return frame_id;
}

// returns first active port_nr in port_mask and clears this port_nr in the mask
U8_CHAR get_port_nr_from_tx_port_mask(void)
{
	U8_CHAR port_mask, n, mask;
	U8_CHAR idx;

	idx = Node.tx_framebuf.tx_rd_ptr;
	port_mask = Node.tx_framebuf.tx_frame[idx].port_mask;
	for(n = 0; n < 8; n++) {
		mask = (1 << n);
		if((port_mask & mask) > 0) {
			// delete bit in mask
			Node.tx_framebuf.tx_frame[idx].port_mask &= ~(mask);
			// return port_nr
			return n;
		}
	}

	return 0;
}

// increments rd pointer of tx buffer
void inc_tx_rd_ptr(void)
{
	Node.tx_framebuf.tx_rd_ptr++;
	if(Node.tx_framebuf.tx_rd_ptr >= NODE_TX_BUFFER_SIZE) Node.tx_framebuf.tx_rd_ptr = 0;
}

// increments wr pointer of tx buffer
void inc_tx_wr_ptr(void)
{
	Node.tx_framebuf.tx_wr_ptr++;
	if(Node.tx_framebuf.tx_wr_ptr >= NODE_TX_BUFFER_SIZE) Node.tx_framebuf.tx_wr_ptr = 0;
	if(Node.tx_framebuf.tx_wr_ptr == Node.tx_framebuf.tx_rd_ptr) Node.dbg_cnt.tx_ovf_cnt++;
}

// check if a package is in tx framebuf
// and if so, send the package to one port
void check_tx_buffer_and_send(void)
{
	U8_CHAR port_nr;

	// check if something in tx framebuffer
	if(UB_Xnet_L1_Tx_Buf_Cnt() > 0) {
		// check if this package has to be send elsewhere
		if(get_tx_port_mask() > 0) {
			// get port to send
			port_nr = get_port_nr_from_tx_port_mask();
			// send package
			if(send_package(port_nr) == TX_OK) {
				if(get_tx_frame_id() != XNET_FRAME_HELLO) {
					// some package was send, so we can skip own next HELLO package to this port
					Node.port[port_nr].hello_cnt = 0;
					Node.port[port_nr].hello_cnt_delay = NODE_HELLO_DELAY_SKIP;
				}
				// check if this package has to be send to another port
				if(get_tx_port_mask() == 0) {
					// no, so we can switch to the next frame
					inc_tx_rd_ptr();
				}
			}
		}
		else {
			// switch to next frame in tx framebuffer
			inc_tx_rd_ptr();
		}
	}
}

// returns number of packages in rx framebuf (0 = empty)
U8_CHAR UB_Xnet_L1_Rx_Buf_Cnt(void)
{
	U8_CHAR cnt = 0;

	if(Node.rx_framebuf.rx_wr_ptr >= Node.rx_framebuf.rx_rd_ptr) {
		cnt = Node.rx_framebuf.rx_wr_ptr - Node.rx_framebuf.rx_rd_ptr;
	}
	else {
		cnt = (Node.rx_framebuf.rx_wr_ptr + NODE_RX_BUFFER_SIZE) - Node.rx_framebuf.rx_rd_ptr;
	}

	return cnt;
}

// returns port_nr of current package in rx buffer
U8_CHAR get_rx_port_nr(void)
{
	U8_CHAR idx, port_nr;

	idx = Node.rx_framebuf.rx_rd_ptr;
	port_nr = Node.rx_framebuf.rx_frame[idx].port_nr;

	return port_nr;
}



// increments rd pointer of rx buffer
void inc_rx_rd_ptr(void)
{
	Node.rx_framebuf.rx_rd_ptr++;
	if(Node.rx_framebuf.rx_rd_ptr >= NODE_RX_BUFFER_SIZE) Node.rx_framebuf.rx_rd_ptr = 0;
}

// increments wr pointer of rx buffer
void inc_rx_wr_ptr(void)
{
	Node.rx_framebuf.rx_wr_ptr++;
	if(Node.rx_framebuf.rx_wr_ptr >= NODE_RX_BUFFER_SIZE) Node.rx_framebuf.rx_wr_ptr = 0;
	if(Node.rx_framebuf.rx_wr_ptr == Node.rx_framebuf.rx_rd_ptr) Node.dbg_cnt.rx_ovf_cnt++;
}



// send a package to an port
// return:
// TX_OK = package has send to neighbor
// TX_NO_RESPONSE = timeout, no response from neighbor
// TX_PACKAGE_SKIPPED = package skipped
// TX_ERROR = error detection from neighbor
TX_State_t send_package(U8_CHAR port_nr)
{
	TX_State_t state;
	U8_CHAR n;

	if(port_nr >= NODE_PORT_CNT) return TX_DATA_ERROR;
	if(Node.port[port_nr].enable == XFALSE) {
		change_tx_connection_state(port_nr, XFALSE);
		return TX_PORT_DISABLED;
	}
	if(get_tx_frame_id() != XNET_FRAME_HELLO) {
		// skip sending to a closed port if package is not a HELLO
		if(Node.port[port_nr].connected == XFALSE) return TX_PACKAGE_SKIPPED;
	}

	// multiple send attempts
	for(n = 0; n < XNET_PACKAGE_TX_RETRY; n++) {
		state = tx_package(port_nr);
		if(state == TX_OK) {
			// neighbor has received and checked frame
			change_tx_connection_state(port_nr, XTRUE);
			return TX_OK;
		}
		else if(state == TX_NO_RESPONSE) {
			// when neighbor is not responding, escape open attempts
			break; // exit for loop
		}
	}

	// sending failed, port is not connected
	change_tx_connection_state(port_nr, XFALSE);

	// increment error counter
	if(state != TX_NO_RESPONSE) {
		Node.dbg_cnt.tx_err_cnt++;
	}

	// error
	return state;
}



// try to send a package to a neighbor
// return:
// TX_OK = package has send to neighbor
// TX_NO_RESPONSE = timeout, no response from neighbor
// TX_ERROR = error detection from neighbor
TX_State_t tx_package(U8_CHAR port_nr)
{
	TX_State_t state;

	// open port
	state = tx_open(port_nr);
	if(state == TX_OK) {
		// send frame
		state = tx_frame(port_nr);
	}
	tx_close(port_nr);

	return state;
}


// try to open port for sending
// return:
// TX_OK = port is open for sending frame
// TX_ERROR = error on open attempt
// TX_NO_RESPONSE = timeout, no response from neighbor
TX_State_t tx_open(U8_CHAR port_nr)
{
	U8_CHAR tx_open_cnt = 0, rx_data;
	TX_State_t state;
	U8_BOOL response = XFALSE;
	U8_BOOL collision = XFALSE;
	U8_CHAR collision_delay = 0;
	U8_CHAR collision_cnt = 0;

	// first check if any neighbor is trying to send too
	check_all_ports_for_rx_requests();

	// multiple open attempts
	while(tx_open_cnt < XNET_OPEN_TX_RETRY) {
		// send port open request
		UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_OPEN_REQ);

		// wait until receiving an port open acknowledge/request or timeout
		Node.cnt.delay_cnt = 0;
		while(Node.cnt.delay_cnt < XNET_OPEN_REQUEST_TIMEOUT_ms) {
			state = tx_check_neighbor_response(port_nr);
			if(state == TX_OK) {
				// open acknowledge received
				response = XTRUE;
				break; // exit timeout
			}
			else if(state == TX_COLLISION) {
				// neighbor want's to send a frame too
				collision = XTRUE;
				collision_cnt++;
				UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_OPEN_COLLISION);
				break; // exit timeout
			}
		}

		if(response == XTRUE) {
			break; // exit while, ok
		}
		else if(collision_cnt > XNET_OPEN_COLLISION_RETRY) {
			Node.dbg_cnt.collision_cnt++;
			break; // exit while, error
		}
		else if(collision == XTRUE) {
			// collision detection
			collision = XFALSE;

			// try to dissolve the collision by different delay time as neighbor
			if((tx_open_cnt & 0x01) == 0) {
				collision_delay = (Node.my_id & 0x0F);
			}
			else {
				collision_delay = ((Node.my_id & 0xF0) >> 4);
			}

			// wait until receiving an port open request or collision timeout
			Node.cnt.delay_cnt = 0;
			while(Node.cnt.delay_cnt < collision_delay) {
				state = tx_check_neighbor_response(port_nr);
				if(state == TX_COLLISION) {
					// neighbor want's to send his frame first
					if(rx_package(port_nr) == RX_OK) {
						// frame received
						// toDo: add something to process rx frame
					}
				}
			}
			// don't increment tx_open_cnt here to send another tx open request
		}
		else {
			tx_open_cnt++;
		}
	}

	// check if neighbor has responded
	if(response == XTRUE) {
		// send tx port open status
		UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_TX_OPEN);

		// wait until rx port open status or timeout
		Node.cnt.delay_cnt = 0;
		while(Node.cnt.delay_cnt < XNET_BYTES_TIMEOUT_ms) {
			if(UB_Xnet_L0_Check_Rx_Buf(port_nr) == true) {
				rx_data = UB_Xnet_L0_Read_Rx_Buf(port_nr);
				if(rx_data == XNET_CMD_RX_OPEN) {
					// neighbor has opened his port to receive
					return TX_OK;
				}
			}
		}
	}
	else {
		return TX_NO_RESPONSE;
	}

	return TX_ERROR;
}

// try to send frame to neighbor
// return:
// TX_OK = sending frame successful
// TX_NO_RESPONSE = timeout, no response from neighbor
// TX_ERROR = error detection from neighbor
TX_State_t tx_frame(U8_CHAR port_nr)
{
	U8_CHAR length, rx_data;
	U8_CHAR idx;

	idx = Node.tx_framebuf.tx_rd_ptr;
	length = Node.tx_framebuf.tx_frame[idx].frame[0];

	// send complete frame to neighbor
	UB_Xnet_L0_Send_Array(port_nr, Node.tx_framebuf.tx_frame[idx].frame, length);

	// wait until frame ok/error or timeout
	Node.cnt.delay_cnt = 0;
	while(Node.cnt.delay_cnt < XNET_BYTES_TIMEOUT_ms) {
		if(UB_Xnet_L0_Check_Rx_Buf(port_nr) == XTRUE) {
			rx_data = UB_Xnet_L0_Read_Rx_Buf(port_nr);
			if(rx_data == XNET_CMD_FRAME_OK) {
				// neighbor has received the frame
				return TX_OK;
			}
			else if(rx_data == XNET_CMD_FRAME_ERROR) {
				// frame error detection
				return TX_ERROR;
			}
		}
	}

	return TX_NO_RESPONSE;
}

// close port after sending
void tx_close(U8_CHAR port_nr)
{
	// send tx port closed
	UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_TX_CLOSED);
}

// check all ports and handle open requests from neighbors
// return response from active port
// return:
// TX_OK = open acknowledge from neighbor received
// TX_NO_RESPONSE = no response from neighbor
// TX_COLLISION = neighbor want's to send too
TX_State_t tx_check_neighbor_response(U8_CHAR port_nr)
{
	TX_State_t state = TX_NO_RESPONSE;
	U8_CHAR n, rx_data;

	// check if open request received on any port
	for(n = 0; n < NODE_PORT_CNT; n++) {
		if(Node.port[n].enable == XTRUE) {
			if(UB_Xnet_L0_Check_Rx_Buf(n) == XTRUE) {
				rx_data = UB_Xnet_L0_Read_Rx_Buf(n);
				if(port_nr != n) {
					// neighbor on another port (handle only open requests)
					if(rx_data == XNET_CMD_OPEN_REQ) {
						// open request received
						if(rx_package(n) == RX_OK) {
							// frame received
							// toDo: add something to process rx frame
						}
					}
				}
				else {
					// neighbor on same port (handle open requests and open acknowledge)
					if(rx_data == XNET_CMD_OPEN_REQ) {
						// neighbor wants to send too
						state = TX_COLLISION;
					}
					else if(rx_data == XNET_CMD_OPEN_ACK) {
						// open acknowledge from neighbor
						state = TX_OK;
					}
				}
			}
		}
	}

	return state;
}

// try to receive a package from a neighbor
// return:
// RX_OK = package has received from neighbor
// RX_NO_RESPONSE = timeout, no response from neighbor
// RX_ERROR = frame error detection
RX_State_t rx_package(U8_CHAR port_nr)
{
	RX_State_t state;

	state = rx_open(port_nr);
	if(state == RX_OK) {
		state = rx_frame(port_nr);
		if(state == RX_OK) {
			// new frame in rx buffer
			change_rx_connection_state(port_nr, XTRUE);

			inc_rx_wr_ptr();
			if(Node.port[port_nr].connected == XTRUE) {
				// package received, so we can skip own hello package (if connection is allready established)
				Node.port[port_nr].hello_cnt = 0;
				Node.port[port_nr].hello_cnt_delay = NODE_HELLO_DELAY_SKIP;
			}
		}
	}
	rx_close(port_nr);

	if(state != RX_OK) {
		change_rx_connection_state(port_nr, XFALSE);
		// increment error counter
		Node.dbg_cnt.rx_err_cnt++;
	}

	return state;
}

// try to open port for receiving
// return:
// RX_OK = port is open for receiving frame
// RX_NO_RESPONSE = timeout, no response from neighbor
RX_State_t rx_open(U8_CHAR port_nr)
{
	U8_CHAR rx_data;
	U8_BOOL response = XFALSE;

	// send port open acknowledge
	UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_OPEN_ACK);

	// wait until tx port open status or timeout
	Node.cnt.delay_cnt = 0;
	while(Node.cnt.delay_cnt < XNET_BYTES_TIMEOUT_ms) {
		if(UB_Xnet_L0_Check_Rx_Buf(port_nr) == XTRUE) {
			rx_data = UB_Xnet_L0_Read_Rx_Buf(port_nr);
			if(rx_data == XNET_CMD_TX_OPEN) {
				// neighbor has opened his port to send
				response = XTRUE;
				break; // exit timeout
			}
			else if(rx_data == XNET_CMD_OPEN_REQ) {
				// neighbor tries to open his port again

				// send port open acknowledge
				UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_OPEN_ACK);
				Node.cnt.delay_cnt = 0;
			}
		}
	}

	// check if neighbor has responded
	if(response == XTRUE) {
		// clear ringbuffer to receive frame from neighbor
		UB_Xnet_L0_Clear_Rx_Ringbuffer(port_nr);
		// send rx port open status
		UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_RX_OPEN);
		return RX_OK;
	}

	return RX_NO_RESPONSE;
}

// try to receive frame from neighbor
// return:
// RX_OK = receiving frame successful
// RX_NO_RESPONSE = timeout, no response from neighbor
// RX_ERROR = frame error detection
RX_State_t rx_frame(U8_CHAR port_nr)
{
	U8_CHAR n, actual_rx_bytes = 0, old_rx_bytes = 0;
	U8_CHAR length = 0;
	U8_BOOL response = XFALSE;
	U8_CHAR idx;

	// delete neighbor NodeId
	Node.port[port_nr].neighbor_id = XNET_NODE_ID_UNDEFINED;

	idx = Node.rx_framebuf.rx_wr_ptr;

	// receiving length of frame
	Node.cnt.delay_cnt = 0;
	while(Node.cnt.delay_cnt < XNET_BYTES_TIMEOUT_ms) {
		actual_rx_bytes = UB_Xnet_L0_Get_Rx_Ringbuffer_Cnt(port_nr);
		if(actual_rx_bytes > 0) {
			// first receive frame length
			length = UB_Xnet_L0_Read_Rx_Ringbuffer(port_nr);
			response = XTRUE;
			break; // exit timeout
		}
	}

	// check if neighbor has responded
	if(response == XFALSE) {
		// send error status
		UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_FRAME_ERROR);
		return RX_NO_RESPONSE;
	}

	// check if length is valid
	if((length < XNET_MIN_FRAME_LENGTH) || (length > XNET_MAX_FRAME_LENGTH)) {
		// send error status
		UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_FRAME_ERROR);
		return RX_ERROR;
	}

	// store port_nr and length information
	Node.rx_framebuf.rx_frame[idx].port_nr = port_nr;
	Node.rx_framebuf.rx_frame[idx].frame[0] = length;

	// receiving rest of frame until timeout or length is ok
	Node.cnt.delay_cnt = 0;
	while(Node.cnt.delay_cnt < XNET_BYTES_TIMEOUT_ms) {
		actual_rx_bytes = UB_Xnet_L0_Get_Rx_Ringbuffer_Cnt(port_nr);
		if(actual_rx_bytes > old_rx_bytes) {
			old_rx_bytes = actual_rx_bytes;
			Node.cnt.delay_cnt = 0;
			if(actual_rx_bytes == (length-1)) break; // exit timeout
		}
	}

	// check length of received frame
	if(actual_rx_bytes != (length-1)) {
		// send error status
		UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_FRAME_ERROR);
		return RX_ERROR;
	}

	// store complete frame
	for(n = 1; n < length; n++) {
		Node.rx_framebuf.rx_frame[idx].frame[n] = UB_Xnet_L0_Read_Rx_Ringbuffer(port_nr);
	}

	// check frame
	if(rx_check_frame(idx) == XFALSE) {
		// send error status
		UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_FRAME_ERROR);
		return RX_ERROR;
	}

	// store neighbor ID
	Node.port[port_nr].neighbor_id = Node.rx_framebuf.rx_frame[idx].frame[2];

	// send ok status
	UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_FRAME_OK);

	return RX_OK;
}

// close port after receiving
void rx_close(U8_CHAR port_nr)
{
	// send rx port closed
	UB_Xnet_L0_Send_Byte(port_nr, XNET_CMD_RX_CLOSED);
}

// check RX frame
// return:
// XTRUE = rx frame ok
// XFALSE = wrong parameter
U8_BOOL rx_check_frame(U8_CHAR idx)
{
	U8_CHAR length;
	U8_CHAR crc_hi,crc_lo;
	U8_CHAR frame_id, neighbor_id;

	length = Node.rx_framebuf.rx_frame[idx].frame[0]; // LEN
	if((length < XNET_MIN_FRAME_LENGTH) || (length > XNET_MAX_FRAME_LENGTH)) return XFALSE;

	frame_id = Node.rx_framebuf.rx_frame[idx].frame[1]; // FRAME_ID
	if(frame_id == XNET_FRAME_UNDEFINED) return XFALSE;

	neighbor_id = Node.rx_framebuf.rx_frame[idx].frame[2]; // neighbor NODE_ID
	if(neighbor_id == XNET_NODE_ID_UNDEFINED) return XFALSE;

	if(calc_checksum(Node.rx_framebuf.rx_frame[idx].frame, &crc_hi, &crc_lo) == XFALSE) return XFALSE;

	if(crc_hi != Node.rx_framebuf.rx_frame[idx].frame[length-2]) return XFALSE;
	if(crc_lo != Node.rx_framebuf.rx_frame[idx].frame[length-1]) return XFALSE;

	return XTRUE;
}

// process received frame and fill rx_payload
// return:
// RX_OK = rx frame ok
// RX_ERROR = wrong parameter
RX_State_t rx_process_frame(U8_CHAR port_nr)
{
	U8_CHAR frame_id;
	U8_CHAR neighbor_id;
	U8_CHAR idx, n;
	U8_CHAR rx_len;

	idx = Node.rx_framebuf.rx_rd_ptr;
	frame_id = Node.rx_framebuf.rx_frame[idx].frame[1];		// frame ID
	neighbor_id = Node.rx_framebuf.rx_frame[idx].frame[2]; // neighbor node ID

	// check ID
	if((neighbor_id == XNET_NODE_ID_UNDEFINED) || (neighbor_id == Node.my_id)) return RX_ERROR;

	// FloodLinkState special
	if(frame_id == XNET_FRAME_FLOOD_LINK_STATE) {
		if(Node.state == NODE_STATE_CONNECTION) Node.cnt.state_cnt = 0;
		Node.dbg_cnt.rec_ls++;
	}

	rx_len = Node.rx_framebuf.rx_frame[idx].frame[0];
	// fill header of the package
	Node.rx_payload.port_nr = port_nr;
	Node.rx_payload.frame_id = frame_id;
	Node.rx_payload.host_id = Node.rx_framebuf.rx_frame[idx].frame[3];
	// fill payload of the package
	Node.rx_payload.data_length = rx_len-6;
	if(Node.rx_payload.data_length > 0) {
		for(n = 0; n < Node.rx_payload.data_length; n++) {
			Node.rx_payload.data[n] = Node.rx_framebuf.rx_frame[idx].frame[n+4];
		}
	}

	// fill debug list
	if(frame_id != XNET_FRAME_HELLO) {
		UB_Xnet_L1_AddDbgLst(DBG_LIST_RX_FRAME, frame_id, neighbor_id, port_nr, Node.rx_payload.host_id, Node.rx_payload.data[0], Node.rx_payload.data[1], Node.rx_payload.data[2], Node.rx_payload.data[3], Node.rx_payload.data[4]);
	}

	return RX_OK;
}

// calculate crc16 checksum of package
// polynom: 0x1021, startvalue: 0x0000 (CRC16_CCIT_ZERO)
// return:
// XTRUE = calculation ok
// XFALSE = wrong parameter
U8_BOOL calc_checksum(U8_CHAR *data_ptr, U8_CHAR *crc_hi, U8_CHAR *crc_lo)
{
	U8_CHAR n, length, data;
	U16_WORD crc = 0x00; // init value for CRC

	length = data_ptr[0];
	if((length < XNET_MIN_FRAME_LENGTH) || (length > XNET_MAX_FRAME_LENGTH)) return XFALSE;

	for(n = 0; n < (length - 2); n++) {
		data = data_ptr[n];
		crc  = (U8_CHAR)(crc >> 8) | (crc << 8);
		crc ^= data;
		crc ^= (U8_CHAR)(crc & 0x00FF) >> 4;
		crc ^= (crc << 8) << 4;
		crc ^= ((crc & 0x00FF) << 4) << 1;
	}

	*crc_hi = (U8_CHAR)(crc >> 8);
	*crc_lo = (U8_CHAR)(crc & 0x00FF);

	return XTRUE;
}

// check all ports and handle open requests from neighbors
void check_all_ports_for_rx_requests(void)
{
	U8_CHAR n, rx_data;

	// check if open request received on any port
	for(n = 0; n < NODE_PORT_CNT; n++) {
		if(Node.port[n].enable == XTRUE) {
			if(UB_Xnet_L0_Check_Rx_Buf(n) == XTRUE) {
				rx_data = UB_Xnet_L0_Read_Rx_Buf(n);
				if(rx_data == XNET_CMD_OPEN_REQ) {
					// open request received
					if(rx_package(n) == RX_OK) {
						// frame received
						// toDo: add something to process rx frame
					}
				}
			}
		}
	}
}

// calculation new connection state of a port
// (result of tx_state and rx_state)
void con_state_changed(U8_CHAR port_nr)
{
	if((Node.port[port_nr].tx_ok == XFALSE) || (Node.port[port_nr].rx_ok == XFALSE)) {
		// tx OR rx is not ok --> connection is lost
		if(Node.port[port_nr].connected != XFALSE) {
			Node.port[port_nr].connected = XFALSE;
			Node.port[port_nr].neighbor_id = XNET_NODE_ID_UNDEFINED;
			// set flag to trigger own link state update
			Node.l2_flags.flags |= L2_FLAG_OWN_LINK_STATE_CHANGED;
			Node.l2_flags.OWN_LINK_STATE_CHANGED_cnt = 0;
			UB_Xnet_L1_AddDbgLst(DBG_LIST_UPD_LS, 0, 0, port_nr, 0, 0, 0, 0, 0, 0);
		}
	}
	else {
		// tx AND rx is ok --> connection is established
		if(Node.port[port_nr].connected != XTRUE) {
			Node.port[port_nr].connected = XTRUE;
			Node.port[port_nr].new_neighbor = XTRUE;
			// set flag to trigger own link state update
			Node.l2_flags.flags |= L2_FLAG_OWN_LINK_STATE_CHANGED;
			Node.l2_flags.OWN_LINK_STATE_CHANGED_cnt = 0;
			UB_Xnet_L1_AddDbgLst(DBG_LIST_UPD_LS, 1, Node.port[port_nr].neighbor_id, port_nr, 0, 0, 0, 0, 0, 0);
		}
	}
}

// change the tx connection state of an port
// (tx state of the port can be connected or disconnected)
void change_tx_connection_state(U8_CHAR port_nr, U8_BOOL new_state)
{
	U8_BOOL old_tx, old_rx;

	old_tx = Node.port[port_nr].tx_ok;
	old_rx = Node.port[port_nr].rx_ok;

	if(new_state == XFALSE) {
		Node.port[port_nr].tx_ok = XFALSE;
		Node.port[port_nr].rx_ok = XFALSE;
	}
	else {
		Node.port[port_nr].tx_ok = XTRUE;
	}

	if((old_tx != Node.port[port_nr].tx_ok) || (old_rx != Node.port[port_nr].rx_ok)) {
		con_state_changed(port_nr);
	}
}

// change the rx connection state of an port
// (rx state of the port can be connected or disconnected)
void change_rx_connection_state(U8_CHAR port_nr, U8_BOOL new_state)
{
	U8_BOOL old_tx, old_rx;

	old_tx = Node.port[port_nr].tx_ok;
	old_rx = Node.port[port_nr].rx_ok;

	if(new_state == XFALSE) {
		Node.port[port_nr].tx_ok = XFALSE;
		Node.port[port_nr].rx_ok = XFALSE;
	}
	else {
		Node.port[port_nr].rx_ok = XTRUE;
	}

	if((old_tx != Node.port[port_nr].tx_ok) || (old_rx != Node.port[port_nr].rx_ok)) {
		con_state_changed(port_nr);
	}
}
